describe 'Given a jQuery element', ->

    beforeEach ->
        element = (@element = $('<a></a>'))

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when initializing the element', ->

        beforeEach ->
            @result = @element.hijax()

        it 'should not return null', ->
            expect(@result).not.toBeNull()

        it 'should return a reference to jQuery', ->
            expect(@result.jquery).toEqual '1.7.2'

    describe 'when hijaxing is invoked', ->

        hits = 0

        beforeEach ->
            hits = 0
            @element.hijax(url: 'some url').click -> hits++

        beforeEach -> # because of
            @element.click() # default trigger

        it 'should respond to the default trigger', ->
            expect($.ajax.calledOnce).toBeTruthy()

        it 'should allow other configured listeners to respond to the same trigger', ->
            expect(hits).toEqual 1

    describe 'when configuring hijax multiple times', ->

        beforeEach ->
            @element.hijax(url: '/first/').hijax(url: '/second')

        beforeEach ->
            @element.click()

        it 'should make a single ajax call', ->
            expect($.ajax.calledOnce).toBeTruthy()

        it 'should use the configuration options from the first init', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/first/'

describe 'Given a hijax call straight off the jQuery object', ->

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    beforeEach ->
        $.hijax({url: '/' })
        $.hijax({url: '/' })

    it 'should immediately hijax against the document element without an external trigger', ->
        expect($.ajax.getCall(0).args[0].url).toEqual '/'

    it 'should should allow multiple, concurrent requests', ->
        expect($.ajax.callCount).toEqual 2

describe 'Given a selector', ->

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when no elements match the selector', ->

        beforeEach ->
            $('#myelement a').hijax
                url: '/first/'

        beforeEach ->
            $(document.body).append('<div id="myelement"><a></a></div>')
            $('#myelement a').click()

        afterEach ->
            $(document).remove('#myelement')

        it 'should still register future (not-yet-created) elemments to be hijaxed', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/first/'

    describe 'when existing elements match the selector', ->

        beforeEach ->
            $(document.body).append('<div id="myelement"><a></a></div>')
            $('#myelement a').hijax(url: '/first/').click()

        afterEach ->
            $(document).remove('#myelement')

        it 'should still register future (not-yet-created) elemments to be hijaxed', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/first/'

    describe 'when the selector captures a form (and no trigger has been explicitly configured)', ->

        element = undefined;

        beforeEach ->
            $(document.body).append('<form class="form-selected" action="/first/"></form>')
            element = $('.form-selected').hijax()

        afterEach ->
            $(document).remove('#myelement')

        it 'should only NOT respond to "click" trigger', ->
            element.click()
            expect($.ajax.called).toBeFalsy()

        it 'should only respond to "submit" trigger', ->
            element.submit();
            expect($.ajax.called).toBeTruthy()

describe 'Given only a url configuration value', ->

    beforeEach ->
        @element = $('<a></a>').hijax
            url: 'my_url'

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when hijaxing is invoked', ->

        beforeEach ->
            @element.click()
            @result = $.ajax.getCall(0).args[0]

        it 'should respond to the default trigger', ->
            expect($.ajax.calledOnce).toBeTruthy()

        it 'should use the configured url', ->
            expect(@result.url).toEqual 'my_url'

        it 'should use the default HTTP method', ->
            expect(@result.type).toEqual 'GET'

        it 'should use the default timeout', ->
            expect(@result.timeout).toEqual 12000

        it 'should use the default synchronicity', ->
            expect(@result.async).toEqual true

        it 'should use the default caching strategy', ->
            expect(@result.cache).toEqual false

        it 'should use the default content type', ->
            expect(@result.contentType).toEqual 'application/json'

        it 'should use the default data type', ->
            expect(@result.dataType).toEqual 'text json'

describe 'Given no url whatsoever', ->

    beforeEach ->
        @element = $('<a></a>').hijax()

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when hijaxing is invoked', ->

        beforeEach ->
            @element.click()

        it 'should NOT call the underlying ajax behavior', ->
            expect($.ajax.called).toBeFalsy()

describe 'Given a zero maximum number of attempts', ->

    beforeEach ->
        @element = $('<a href="/"></a>').hijax(attempts: 0).click()

    describe 'when a request is made', ->

        it 'should not leave the originating element "locked"', ->
            expect(@element[0].hijaxId).not.toBeDefined()

describe 'Given a concurrent configuration setting', ->

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    beforeEach ->
        @element = $('<a href="/"></a>').hijax(concurrent: true).click().click()

    describe 'when multiple requests are made request is made', ->

        it 'should not "lock" the element in question', ->
            expect(@element[0].hijaxId).not.toBeDefined()

        it 'should invoke the hijax behavior multiple times', ->
            expect($.ajax.calledTwice).toBeTruthy()

describe 'Given a set of configuration options', ->

    beforeEach ->
        @element = $('<a></a>').hijax @options =
            trigger: 'my-trigger'
            url: 'my_url'
            method: 'my_method'
            timeout: 1234
            async: false
            cache: true
            headers: { first: 'a', second: 'b' }
            contentType: 'application/json'
            dataType: 'text'

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when hijaxing is invoked', ->

        beforeEach ->
            @element.trigger('my-trigger')
            @result = $.ajax.getCall(0).args[0]

        it 'should respond to the configured trigger', ->
            expect($.ajax.calledOnce).toBeTruthy()

        it 'should use the configured url', ->
            expect(@result.url).toEqual 'my_url'

        it 'should append the configured headers to the request', ->
            expect(@result.headers['first']).toEqual 'a'
            expect(@result.headers['second']).toEqual 'b'

        it 'should use the the configured HTTP method', ->
            expect(@result.type).toEqual 'my_method'

        it 'should use the configured timeout', ->
            expect(@result.timeout).toEqual @options.timeout

        it 'should use the configured synchronicity', ->
            expect(@result.async).toEqual @options.async

        it 'should use the configured caching strategy', ->
            expect(@result.cache).toEqual @options.cache

        it 'should use the configured content type', ->
            expect(@result.contentType).toEqual @options.contentType

        it 'should use the configured data type', ->
            expect(@result.dataType).toEqual @options.dataType

    describe 'when the default trigger is invoked', ->

        beforeEach ->
            @element.click()

        it 'should NOT respond to the default trigger', ->
            expect($.ajax.calledOnce).toBeFalsy()

describe 'Given a configured "before" callback', ->

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when the configured "before" callback is invoked', ->

        context = undefined
        event = undefined

        beforeEach ->
            @element = $('<a href="/page/"></a>').hijax
                before: (e) ->
                    context = this
                    event = e

        beforeEach ->
            @element.click()

        it 'should use the jQuery element as the context', ->
            expect(context.is('a')).toEqual true

        it 'should provide the event that triggered the behavior', ->
            expect(event.type).toEqual 'click'

    describe 'when the configured "before" callback explictly returns false', ->

        beforeEach ->
            @element = $('<custom href="/page/"></custom>').hijax
                before: (e) -> false
            @element.click()

        it 'should not invoke any ajax behavior', ->
            expect($.ajax.called).toBeFalsy()

    describe 'when the configured "before" callback returns a value other than false (including undefined or null)', ->

        beforeEach ->
            @element = $('<custom href="/page/"></custom>').hijax
                before: (e) -> undefined
            @element.click()

        it 'should invoke the expected ajax behavior', ->
            expect($.ajax.called).toBeTruthy()

describe 'Given a configured "progress" callback', ->

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()
    
    describe 'when the configured "progress" callback is invoked', ->
    
        context = undefined
        attempt = undefined
        state = undefined

        beforeEach ->
            @element = $('<a href="/page/"></a>').hijax
                before: ->
                    "my state"
                progress: (_attempt, _state) ->
                    context = this
                    attempt = _attempt
                    state = _state

        beforeEach ->
            @element.click()

        it 'should use the jQuery element as the context', ->
            expect(context.is('a')).toEqual true
            
        it 'should indicate which attempt is being invoked', ->
            expect(attempt).toEqual 1

        it 'should provide the state as a parameter', ->
            expect(state).toEqual "my state"
    
    describe 'when the configured "progress" callback explictly returns false', ->

        completed = undefined

        beforeEach ->
            @element = $('<a href="/"></a>').hijax
                progress: -> false
                complete: -> completed = true
            @element.click()

        it 'should not attempt any hijaxing', ->
            expect($.ajax.called).toBeFalsy()

        it 'should invoke the configured "complete" callback', ->
            expect(completed).toBeTruthy()

describe 'Given a request built at invoke', ->

    request = null
    configuredUrl = null
    context = undefined
    state = undefined
    config = undefined
    

    beforeEach ->
        request =
            url: 'request-url',
            data: { key: 'value1' },
            headers: { configured: 'overwritten', 'request':'custom' }
            method: 'request-method'
            timeout: 0 # forever
            async: true
            cache: false
            contentType: 'request/type'
            dataType: 'request defined'

        @element = $('<a></a>').hijax @options =
            trigger: 'my-trigger'
            url: 'configured-url'
            method: 'configured-method'
            timeout: 1234
            async: false
            cache: true,
            headers: { configured: 'default' },
            contentType: 'application/json'
            dataType: 'text',
            before: ->
                "my state"
            request: (url, _state, _config) ->
                context = this
                configuredUrl = url
                state = _state
                config = _config
                request

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when hijaxing is invoked', ->

        beforeEach ->
            @element.trigger('my-trigger')
            @result = $.ajax.getCall(0).args[0]

        it 'should respond to the configured trigger', ->
            expect($.ajax.calledOnce).toBeTruthy()
        
        it 'should use the jQuery element as the context', ->
            expect(context.is('a')).toEqual true
            
        it 'should provide the configuration url to the request callback', ->
            expect(@options.url).toEqual configuredUrl

        it 'should provide the state from the "before" callback as a parameter', ->
            expect(state).toEqual "my state"
        
        it 'should provide the configuration values as a parameter', ->
            expect(config.trigger).toEqual @options.trigger

        it 'should use the request url over the configured value', ->
            expect(@result.url).toEqual request.url

        it 'should use the request-specific header over the configured value', ->
            expect(@result.headers['configured']).toEqual 'overwritten'

        it 'should append the request header to the request payload', ->
            expect(@result.headers['request']).toEqual 'custom'

        it 'should stringify any payload object given', ->
            expect(@result.data).toEqual JSON.stringify(request.data)

        it 'should use the the request-specific HTTP method over the configured value', ->
            expect(@result.type).toEqual request.method

        it 'should use the request timeout over the configured value', ->
            expect(@result.timeout).toEqual request.timeout

        it 'should use the request synchronicity over the configured value', ->
            expect(@result.async).toEqual request.async

        it 'should use the request caching strategy over the configured value', ->
            expect(@result.cache).toEqual request.cache

        it 'should use the request content type over the configured value', ->
            expect(@result.contentType).toEqual request.contentType

        it 'should use the request data type over the configured value', ->
            expect(@result.dataType).toEqual request.dataType

    describe 'when the request does not specify a url', ->

        beforeEach ->
            request['url'] = undefined
            @element.trigger('my-trigger')
            @result = $.ajax.getCall(0).args[0]

        it 'should use the configured url', ->
            expect(@result.url).toEqual @options.url

describe 'Given global configuration settings', ->

    describe 'when new configuration settings are provided', ->

        beforeEach ->
            $.hijaxSetup(url: '/')
            $.hijaxSetup() # clear previous settings
            $('<a></a>').hijax();

        it 'should overwrite all previous global settings', ->
            expect($.ajax.called).toBeFalsy()

    describe 'when global settings are provided', ->

        beforeEach ->
            $.hijaxSetup
                url: '/global/'
                async: false
                cache: false
                method: 'HEAD'
                timeout: 42
                contentType: 'app/global-settings'
                dataType: 'script'
                headers:
                    key1: 'value1'
                    key2: 'value2'

            $('<a></a>').hijax().click()
            @request = $.ajax.getCall(0).args[0]

        afterEach ->
            $.hijaxSetup() # clear settings

        it 'should use the "url" attribute from the global settings', ->
            expect(@request.url).toEqual '/global/'

        it 'should use the "async" attribute from the global settings', ->
            expect(@request.async).toBeFalsy()

        it 'should use the "method" attribute from the global settings', ->
            expect(@request.type).toEqual 'HEAD'

        it 'should use the "cache" attribute from the global settings', ->
            expect(@request.cache).toBeFalsy()

        it 'should use the "timeout" attribute from the global settings', ->
            expect(@request.timeout).toEqual 42

        it 'should use the "contentType" attribute from the global settings', ->
            expect(@request.contentType).toEqual 'app/global-settings'

        it 'should use the "dataType" attribute from the global settings', ->
            expect(@request.dataType).toEqual 'script'

        it 'should use append all configured headers from the global settings', ->
            expect(@request.headers.key1).toEqual 'value1'
            expect(@request.headers.key2).toEqual 'value2'

    describe 'when hijax options are provided', ->

        element = undefined

        beforeEach ->
            $.hijaxSetup
                url: '/global/'
                async: false
                cache: false
                method: 'HEAD'
                timeout: 42
                contentType: 'app/global-settings'
                dataType: 'script'
                headers:
                    key1: 'global1'
                    key2: 'global2'

            element = $('<a></a>').hijax
                url: '/local/'
                async: true
                cache: true
                method: 'POST'
                timeout: 43
                contentType: 'app/local-settings'
                dataType: 'text xml'
                headers:
                    key1: 'local1' # overwrite
                    key3: 'local3' # new key

            element.click()
            @request = $.ajax.getCall(0).args[0]

        afterEach ->
            $.hijaxSetup() # clear settings

        xit 'should use the local "url" attribute over the the global settings', ->
            expect(@request.url).toEqual '/local/'

        xit 'should use the local "async" attribute over the global settings', ->
            expect(@request.async).toBeTruthy()

        xit 'should use the local"method" attribute over the global settings', ->
            expect(@request.type).toEqual 'POST'

        xit 'should use the local"cache" attribute over the global settings', ->
            expect(@request.cache).toBeTruthy()

        xit 'should use the local"timeout" attribute over the global settings', ->
            expect(@request.timeout).toEqual 43

        xit 'should use the local"contentType" attribute over the global settings', ->
            expect(@request.contentType).toEqual 'app/local-settings'

        xit 'should use the local "dataType" attribute over the global settings', ->
            expect(@request.dataType).toEqual 'text xml'

        it 'should use append all configured headers from the local settings and overwrite colliding keys', ->
            expect(@request.headers.key1).toEqual 'local1'
            expect(@request.headers.key2).toEqual 'global2'
            expect(@request.headers.key3).toEqual 'local3'

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()
        $.hijaxSetup() # clear previous settings

describe 'Given a form element', ->

    beforeEach ->
        @element = $('<form action="/myaction/" method="get"><input type="hidden" name="key1" value="value1"><input type="hidden" name="key2" value="value2"></form>')

    beforeEach ->
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when the "click" trigger is invoked', ->

        beforeEach ->
            @element.hijax().click()

        it 'should NOT respond', ->
            expect($.ajax.called).toBeFalsy()

    describe 'when the "submit" trigger is invoked', ->

        beforeEach ->
            @element.hijax().submit()

        it 'should serialize the form inputs as the payload', ->
            expect($.ajax.getCall(0).args[0].data).toEqual 'key1=value1&key2=value2'

        it 'should specify the correct Content-Type', ->
            expect($.ajax.getCall(0).args[0].contentType).toEqual 'application/x-www-form-urlencoded'

        it 'should respond', ->
            expect($.ajax.called).toBeTruthy()

    describe 'when no action is specified', ->

        beforeEach ->
            @element.attr('action', null).hijax().submit()

        it 'should not invoke the behavior', ->
            expect($.ajax.called).toBeFalsy()

    describe 'when the configuration and form specify differing HTTP methods', ->

        beforeEach ->
            @element.hijax(method: 'post').submit()

        it 'should use the method on the form', ->
            expect($.ajax.getCall(0).args[0].type).toEqual 'get'

describe 'Given an element', ->

    beforeEach ->
        @element = $('<a></a>')
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when the element contains an "href" attribute', ->

        beforeEach ->
            @element.attr('href', '/element-url/').hijax(url: '/configured-url/').click()

        it 'should use the element "href" value over the configuration value', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/element-url/'

    describe 'when the element contains a "method" attribute', ->

        beforeEach ->
            @element.attr('href', '/element-url/').attr('method', 'element-method').hijax(method: 'configured-method').click()

        it 'should use the element "method" value over the configuration value', ->
            expect($.ajax.getCall(0).args[0].type).toEqual 'element-method'

describe 'Given a set of similar elements', ->

    beforeEach ->
        @elements = $('<a href="/first/"></a><a href="/second/"></a>')
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when invoking the hijax behavior on the set of elements', ->

        beforeEach ->
            @elements.hijax().click()

        it 'should invoke each element', ->
            expect($.ajax.calledTwice).toBeTruthy()

        it 'should invoke each element with its respective "href" value', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/first/'
            expect($.ajax.getCall(1).args[0].url).toEqual '/second/'

describe 'Given a form and hyperlink element', ->

    beforeEach ->
        @elements = $('<form action="/form/"></form><a href="/hyperlink/"></a>')
        @xhr = sinon.useFakeXMLHttpRequest()
        sinon.spy($, "ajax")

    afterEach ->
        @xhr.restore()
        $.ajax.restore()

    describe 'when triggering a "click" on the set of elements', ->

        beforeEach ->
            @elements.hijax().click()

        it 'should invoke only one element', ->
            expect($.ajax.calledOnce).toBeTruthy()

        it 'should invoke the hyperlink element', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/hyperlink/'

    describe 'when triggering a "submit" on the set of elements', ->

        beforeEach ->
            @elements.hijax().submit()

        it 'should invoke only one element', ->
            expect($.ajax.calledOnce).toBeTruthy()

        it 'should invoke the form element', ->
            expect($.ajax.getCall(0).args[0].url).toEqual '/form/'

describe 'Given an immediate hijax request, rather than the typical defered/setup request', ->

    count = 0

    beforeEach ->
        @server = sinon.fakeServerWithClock.create()
        @elements = $('<a href="/dir/"></a>')

    afterEach ->
        @server.restore()

    describe 'when the options provided specify a trigger value of "now"', ->

        it 'should make the request without any external trigger', ->
            @elements.hijax(trigger: 'NOW', complete: -> count++) # case insensitive
            @server.respond()
            expect(count).toEqual 1

        it 'should allow additional immediate hijax invocations (after the server has responded and closed the first request)', ->
            @elements.hijax(trigger: 'now', complete: -> count++)
            @server.respond()
            expect(count).toEqual 2

describe 'Given a valid request', ->

    state = "object state"
    counts = {}
    states = {}
    location = null
    data = null
    xhr1 = null
    xhr2 = null
    server = null
    requests = null
    validation = null

    beforeEach ->
        states = {}
        counts = {}
        location = null
        data = null
        xhr1 = null
        xhr2 = null
        requests = null
        validation = null

        @options =
            attempts: 3
            before: ->
                counts.before = increment(counts.before)
                return state
            request: (url, s) ->
                counts.request = increment(counts.request)
                states.request = s
            progress: (attempt, s) ->
                counts.progress = attempt
                states.progress = s
            success: (xhr, payload, s) ->
                xhr1 = xhr
                counts.success = increment(counts.success)
                data = payload
                states.success = s
            failure: (xhr, s) ->
                xhr1 = xhr
                counts.failure = increment(counts.failure)
                states.failure = s
            complete: (xhr, s) ->
                xhr2 = xhr
                counts.complete = increment(counts.complete)
                states.complete = s
            redirect: (xhr, url, s) ->
                xhr1 = xhr
                location = url
                counts.redirect = increment(counts.redirect)
                states.redirect = s
            validation: (xhr, parsed, s) ->
                xhr1 = xhr
                counts.validation = increment(counts.validation)
                validation = parsed
                states.validation = s
            authentication: (xhr, s) ->
                xhr1 = xhr
                counts.authentication = increment(counts.authentication)
                states.authentication = s
            authorization: (xhr, s) ->
                xhr1 = xhr
                counts.authorization = increment(counts.authorization)
                states.authorization = s

        @element = $('<a href="/"></a>').hijax(@options)

    increment = (original) ->
        (original || 0) + 1

    respondWith = (statusCode, responseText) ->
        server.respondWith([statusCode, { 'Content-Type': 'application/json' }, responseText || 'true' ]);

    beforeEach ->
        server = sinon.fakeServerWithClock.create()
        requests = [];
        server.xhr.onCreate = (xhr) ->
            requests.push(xhr)
            server.addRequest(xhr)

    afterEach ->
        server.restore()

    describe 'when submitting a request', ->

        beforeEach ->
            @isGuid = /^(\{){0,1}[0-9A-F]{8}\-[0-9A-F]{4}\-[0-9A-F]{4}\-[0-9A-F]{4}\-[0-9A-F]{12}(\}){0,1}$/i
            respondWith(200)
            @element.click()
            server.respond()

        it 'should invoke the configured "before" callback', ->
            expect(counts.before).toEqual 1

        it 'should invoke the configured "request" callback', ->
            expect(counts.request).toEqual 1

        it 'should provide the state to the configured "request" callback', ->
            expect(states.request).toEqual state

        it 'should append a value uniquely identifying the request to the request headers', ->
            expect(@isGuid.test(requests[0].requestHeaders['x-hijax-id'])).toBeTruthy()

        it 'should invoke the configured "progress" callback', ->
            expect(counts.progress).toEqual 1

        it 'should provide the state to the configured "progress" callback', ->
            expect(states.progress).toEqual state

    describe 'when submitting multiple requests attached to different submissions', ->

        beforeEach ->
            respondWith(200)
            @element.click() # first submission
            server.respond()
            @element.click() # second submission

        it 'should not use the same value for subsequent requests that are part of a different submission', ->
            expect(requests[0].requestHeaders['x-hijax-id']).not.toEqual(requests[1].requestHeaders['x-hijax-id'])

    describe 'when any non-retry response is received', ->

        beforeEach ->
            respondWith(200)
            @element.click()
            server.respond()

        it 'should invoke the configured "complete" callback', ->
            expect(counts.complete).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "complete" callback', ->
            expect(xhr2.status).toEqual 200

        it 'should provide the state to the configured "complete" callback', ->
            expect(states.complete).toEqual state

    describe 'when the server returns "2xx"', ->

        beforeEach ->
            respondWith(200)
            @element.click()
            server.respond()

        it 'should invoke the configured "success" callback', ->
            expect(counts.success).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "success" callback', ->
            expect(xhr1.status).toEqual 200

        it 'should provide the parsed data as a parameter to the "success" callback', ->
            expect(data).toEqual true # the server responds with 'true'

        it 'should provide the state to the configured "success" callback', ->
            expect(states.success).toEqual state

    describe 'when the server returns "201"', ->

        beforeEach ->
            respondWith(201)
            @element.click()
            server.respond()

        it 'should invoke the configured "success" callback', ->
            expect(counts.success).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "success" callback', ->
            expect(xhr1.status).toEqual 201

        it 'should provide the state to the configured "success" callback', ->
            expect(states.success).toEqual state

    describe 'when the server returns "304"', ->

        beforeEach ->
            respondWith(304)
            @element.click()
            server.respond()

        it 'should invoke the configured "success" callback', ->
            expect(counts.success).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "success" callback', ->
            expect(xhr1.status).toEqual 304

        it 'should provide the state to the configured "success" callback', ->
            expect(states.success).toEqual state

    describe 'when the server response contains a "location" header', ->

        beforeEach ->
            server.respondWith([200, { 'Location': '/my-redirect-url/' }, '{}' ]);
            @element.click()
            server.respond()

        it 'should NOT invoke the configured "success" callback', ->
            expect(counts.success || 0).toEqual 0

        it 'should invoke the configured "redirect" callback', ->
            expect(location).toEqual '/my-redirect-url/'

        it 'should provide the redirect header to the configured "redirect" callback', ->
            expect(counts.redirect).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "redirect" callback', ->
            expect(xhr1.status).toEqual 200

        it 'should provide the state to the configured "redirect" callback', ->
            expect(states.redirect).toEqual state

        it 'should still invoke the configured "complete" callback', ->
            expect(counts.complete).toEqual 1

        it 'should provide the state to the configured "complete" callback', ->
            expect(states.complete).toEqual state

    describe 'when the server returns "400"', ->

        beforeEach ->
            respondWith(400, '{ "customKey": "some value" }')
            @element.click()
            server.respond()

        it 'should NOT invoke the configured "success" callback', ->
            expect(counts.success || 0).toEqual 0

        it 'should invoke the configured "validation" callback', ->
            expect(counts.validation).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "validation" callback', ->
            expect(xhr1.status).toEqual 400

        it 'should provide the state to the configured "validation" callback', ->
            expect(states.validation).toEqual state

        it 'should provide a parsed response to the configured "validation" callback', ->
            expect(validation.customKey).toEqual 'some value'

    describe 'when the server returns "401"', ->

        beforeEach ->
            respondWith(401)
            @element.click()
            server.respond()

        it 'should NOT invoke the configured "success" callback', ->
            expect(counts.success || 0).toEqual 0

        it 'should invoke the configured "authentication" callback', ->
            expect(counts.authentication).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "authentication" callback', ->
            expect(xhr1.status).toEqual 401

        it 'should provide the state to the configured "authentication" callback', ->
            expect(states.authentication).toEqual state

    describe 'when the server returns "403"', ->

        beforeEach ->
            respondWith(403)
            @element.click()
            server.respond()

        it 'should NOT invoke the configured "success" callback', ->
            expect(counts.success || 0).toEqual 0

        it 'should invoke the configured "authorization" callback', ->
            expect(counts.authorization).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "authorization" callback', ->
            expect(xhr1.status).toEqual 403

        it 'should provide the state to the configured "authorization" callback', ->
            expect(states.authorization).toEqual state

    describe 'when the server returns "404"', ->

        beforeEach ->
            respondWith(404)
            @element.click()
            server.respond()

        it 'should NOT invoke the configured "success" callback', ->
            expect(counts.success || 0).toEqual 0

        it 'should only attempt the request once', ->
            expect(counts.progress).toEqual 1

        it 'should invoke the configured "failure" callback', ->
            expect(counts.failure).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "failure" callback', ->
            expect(xhr1.status).toEqual 404

        it 'should provide the state to the configured "failure" callback', ->
            expect(states.failure).toEqual state

    describe 'when the server returns an unrecognized 4/5xx error code-level error code (409, 503, etc.)', ->

        beforeEach ->
            respondWith(503)
            @element.click()
            server.respond()

        it 'should NOT invoke the configured "success" callback', ->
            expect(counts.success || 0).toEqual 0

        it 'should only attempt the request once', ->
            expect(counts.progress).toEqual 1

        it 'should invoke the configured "failure" callback', ->
            expect(counts.failure).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "failure" callback', ->
            expect(xhr1.status).toEqual 503

        it 'should provide the state to the configured "failure" callback', ->
            expect(states.failure).toEqual state

    describe 'when the server returns "500"', ->

        beforeEach ->
            respondWith(500)
            @element.click()
            server.respond()

        it 'should only invoke the configured "before" callback the first time and not during any retry attempts', ->
            expect(counts.before).toEqual 1

        it 'should only invoke the configured "request" callback the first time and not during any retry attempts', ->
            expect(counts.request).toEqual 1

        it 'should try the request again up to the configured number of attempts', ->
            expect(counts.progress).toEqual 3

    describe 'when retrying the submission', ->

        beforeEach ->
            respondWith(500)
            @element.click()
            server.respond()

        it 'should only invoke the configured "before" callback the first time and not during any retry attempts', ->
            expect(counts.before).toEqual 1

        it 'should only invoke the configured "request" callback the first time and not during any retry attempts', ->
            expect(counts.request).toEqual 1

        it 'should invoke the configured "progress" callback for each additional request', ->
            expect(counts.progress).toEqual 3

        it 'should contain the same unique identifier for all retry attempts of the same submission', ->
            expect(requests[0].requestHeaders['x-hijax-id']).toEqual(requests[1].requestHeaders['x-hijax-id'])

    describe 'when giving up the submission after too many failed requests', ->

        beforeEach ->
            respondWith(500)
            @element.click()
            server.respond()

        it 'should invoke the configured "failure" callback', ->
            expect(counts.failure).toEqual 1

        it 'should provide the raw xhr to as a parameter to the "failure" callback', ->
            expect(xhr1.status).toEqual 500

        it 'should provide the state to the configured "failure" callback', ->
            expect(states.failure).toEqual state

        it 'should invoke the configured "complete" callback', ->
            expect(counts.complete).toEqual 1

        it 'should provide the state to the configured "complete" callback', ->
            expect(states.complete).toEqual state

    describe 'when another submission is attempted while the first is processing', ->

        beforeEach ->
            respondWith(200)
            @element.click() # first submission
            @element.click() # second submission should be ignored
            server.respond()

        it 'should NOT invoke the underlying ajax behavior', ->
            expect(counts.complete).toEqual 1

describe 'Given a callback which throws an exception', ->

    server = undefined
    completed = false
    exception = undefined
    exceptionXhr = undefined
    exceptionState = undefined

    beforeEach ->
        completed = false
        @element = $('<a href="/"></a>').hijax
            before: -> 'some state'
            success: -> throw 'some exception'
            complete: -> completed = true
            exception: (xhr, ex, s) ->
                exception = ex
                exceptionXhr = xhr
                exceptionState = s

    beforeEach ->
        server = sinon.fakeServer.create()

    afterEach ->
        server.restore()

    respondWith = (statusCode) ->
        server.respondWith([statusCode, { 'Content-Type': 'application/json' }, '{}' ]);

    describe 'when a request is made', ->

        beforeEach ->
            respondWith(200)
            @element.click()
            server.respond()

        it 'should invoke the "exception" callback', ->
            expect(exception).toEqual 'some exception'

        it 'should provide the raw xhr to as a parameter to the "exception" callback', ->
            expect(exceptionXhr).toBeDefined()

        it 'should provide the state to the configured "exception" callback', ->
            expect(exceptionState).toEqual 'some state'

        it 'should invoke the "completed" callback', ->
            expect(completed).toBeTruthy()

        it 'should remove the hijax ID from the underlying element', ->
            expect(@element.hijaxId).not.toBeDefined()